@@ -1,6 +1,6 @@
 -- |
 -- Module      :  Text.Megaparsec
--- Copyright   :  © 2015–2016 Megaparsec contributors
+-- Copyright   :  © 2015–2017 Megaparsec contributors
 --                © 2007 Paolo Martini
 --                © 1999–2001 Daan Leijen
 -- License     :  FreeBSD
@@ -33,7 +33,15 @@
 -- > -- import Text.Megaparsec.Text.Lazy
 --
 -- As you can see the second import depends on data type you want to use as
--- input stream. It just defines useful type-synonym @Parser@.
+-- input stream. It just defines the useful type-synonym @Parser@.
+--
+-- Megaparsec 5 uses some type-level machinery to provide flexibility
+-- without compromising on type safety. Thus type signatures are sometimes
+-- necessary to avoid ambiguous types. If you're seeing a error message that
+-- reads like “Ambiguous type variable @e0@ arising from … prevents the
+-- constraint @(ErrorComponent e0)@ from being resolved”, you need to give
+-- an explicit signature to your parser to eliminate the ambiguity. It's a
+-- good idea to provide type signatures for all top-level definitions.
 --
 -- Megaparsec is capable of a lot. Apart from this standard functionality
 -- you can parse permutation phrases with "Text.Megaparsec.Perm",
@@ -45,13 +53,13 @@ module Text.Megaparsec
   ( -- * Running parser
     Parsec
   , ParsecT
+  , parse
+  , parseMaybe
+  , parseTest
   , runParser
   , runParser'
   , runParserT
   , runParserT'
-  , parse
-  , parseMaybe
-  , parseTest
     -- * Combinators
   , (A.<|>)
   -- $assocbo
@@ -70,6 +78,7 @@ module Text.Megaparsec
   , lookAhead
   , notFollowedBy
   , withRecovery
+  , observing
   , eof
   , token
   , tokens
@@ -140,6 +149,8 @@ module Text.Megaparsec
   , ShowToken (..)
   , ShowErrorComponent (..)
   , parseErrorPretty
+    -- * Debugging
+  , dbg
     -- * Low-level operations
   , Stream (..)
   , State (..)
@@ -149,6 +160,8 @@ module Text.Megaparsec
   , setPosition
   , pushPosition
   , popPosition
+  , getTokensProcessed
+  , setTokensProcessed
   , getTabWidth
   , setTabWidth
   , getParserState
@@ -177,15 +190,20 @@ import Text.Megaparsec.Prim
 
 -- $many
 --
--- @many p@ applies the parser @p@ /zero/ or more times. Returns a list of
--- the returned values of @p@.
+-- @many p@ applies the parser @p@ /zero/ or more times and returns a list
+-- of the returned values of @p@. Note that if the @p@ parser fails
+-- consuming input, then the entire @many p@ parser fails with the error
+-- message @p@ produced instead of just stopping iterating. In these cases
+-- wrapping @p@ with 'try' may be desirable.
 --
--- > identifier = (:) <$> letter <*> many (alphaNum <|> char '_')
+-- > identifier = (:) <$> letter <*> many (alphaNumChar <|> char '_')
 
 -- $some
 --
--- @some p@ applies the parser @p@ /one/ or more times. Returns a list of
--- the returned values of @p@.
+-- @some p@ applies the parser @p@ /one/ or more times and returns a list of
+-- the returned values of @p@. The note about behavior of the combinator in
+-- case when @p@ fails consuming input (see 'A.many') applies to 'some' as
+-- well.
 --
 -- > word = some letter
 
