@@ -1,6 +1,6 @@
 -- |
 -- Module      :  Text.Megaparsec.Char
--- Copyright   :  © 2015–2016 Megaparsec contributors
+-- Copyright   :  © 2015–2017 Megaparsec contributors
 --                © 2007 Paolo Martini
 --                © 1999–2001 Daan Leijen
 -- License     :  FreeBSD
@@ -402,7 +402,7 @@ string' = tokens casei
 -- | Case-insensitive equality test for characters.
 
 casei :: Char -> Char -> Bool
-casei x y = toLower x == toLower y
+casei x y = toUpper x == toUpper y
 {-# INLINE casei #-}
 
 -- | Case-insensitive 'elem'.
