@@ -1,6 +1,6 @@
 -- |
 -- Module      :  Text.Megaparsec.Error
--- Copyright   :  © 2015–2016 Megaparsec contributors
+-- Copyright   :  © 2015–2017 Megaparsec contributors
 -- License     :  FreeBSD
 --
 -- Maintainer  :  Mark Karpov <markkarpov@opmbx.org>
@@ -26,7 +26,8 @@ module Text.Megaparsec.Error
   , ShowToken (..)
   , ShowErrorComponent (..)
   , parseErrorPretty
-  , sourcePosStackPretty )
+  , sourcePosStackPretty
+  , parseErrorTextPretty )
 where
 
 import Control.DeepSeq
@@ -40,13 +41,14 @@ import Data.Set (Set)
 import Data.Typeable (Typeable)
 import GHC.Generics
 import Prelude hiding (concat)
+import Test.QuickCheck hiding (label)
 import qualified Data.List.NonEmpty as NE
 import qualified Data.Set           as E
 
 import Text.Megaparsec.Pos
 
 #if !MIN_VERSION_base(4,8,0)
-import Control.Applicative ((<$>))
+import Control.Applicative
 #endif
 
 -- | Data type that is used to represent “unexpected\/expected” items in
@@ -62,6 +64,18 @@ data ErrorItem t
 
 instance NFData t => NFData (ErrorItem t)
 
+instance Arbitrary t => Arbitrary (ErrorItem t) where
+  arbitrary = oneof
+    [
+#if !MIN_VERSION_QuickCheck(2,9,0)
+      Tokens <$> (NE.fromList . getNonEmpty <$> arbitrary)
+    , Label  <$> (NE.fromList . getNonEmpty <$> arbitrary)
+#else
+      Tokens <$> arbitrary
+    , Label  <$> arbitrary
+#endif
+    , return EndOfInput ]
+
 -- | The type class defines how to represent information about various
 -- exceptional situations. Data types that are used as custom data component
 -- in 'ParseError' must be instances of this type class.
@@ -103,6 +117,13 @@ instance NFData Dec where
   rnf (DecFail str) = rnf str
   rnf (DecIndentation ord ref act) = ord `seq` rnf ref `seq` rnf act
 
+instance Arbitrary Dec where
+  arbitrary = oneof
+    [ sized (\n -> do
+        k <- choose (0, n `div` 2)
+        DecFail <$> vectorOf k arbitrary)
+    , DecIndentation <$> arbitrary <*> arbitrary <*> arbitrary ]
+
 instance ErrorComponent Dec where
   representFail        = DecFail
   representIndentation = DecIndentation
@@ -139,7 +160,35 @@ instance (Ord t, Ord e) => Monoid (ParseError t e) where
   mappend = (<>)
   {-# INLINE mappend #-}
 
-instance (Show t, Typeable t, Show e, Typeable e) => Exception (ParseError t e)
+instance ( Show t
+         , Typeable t
+         , Ord t
+         , ShowToken t
+         , Show e
+         , Typeable e
+         , ShowErrorComponent e )
+  => Exception (ParseError t e) where
+#if MIN_VERSION_base(4,8,0)
+  displayException = parseErrorPretty
+#endif
+
+instance (Arbitrary t, Ord t, Arbitrary e, Ord e)
+    => Arbitrary (ParseError t e) where
+  arbitrary = ParseError
+#if MIN_VERSION_QuickCheck(2,9,0)
+    <$> arbitrary
+#else
+    <$> (NE.fromList . getNonEmpty <$> arbitrary)
+#endif
+#if MIN_VERSION_QuickCheck(2,8,2)
+    <*> arbitrary
+    <*> arbitrary
+    <*> arbitrary
+#else
+    <*> (E.fromList <$> arbitrary)
+    <*> (E.fromList <$> arbitrary)
+    <*> (E.fromList <$> arbitrary)
+#endif
 
 -- | Merge two error data structures into one joining their collections of
 -- message items and preferring longest match. In other words, earlier error
@@ -186,15 +235,41 @@ stringPretty xs           = "\"" ++ NE.toList xs ++ "\""
 -- character @ch@, suitable for using in error messages.
 
 charPretty :: Char -> String
-charPretty '\0' = "null"
-charPretty '\a' = "bell"
-charPretty '\b' = "backspace"
-charPretty '\t' = "tab"
-charPretty '\n' = "newline"
-charPretty '\v' = "vertical tab"
-charPretty '\f' = "form feed"
-charPretty '\r' = "carriage return"
-charPretty ' '  = "space"
+charPretty '\NUL' = "null (control character)"
+charPretty '\SOH' = "start of heading (control character)"
+charPretty '\STX' = "start of text (control character)"
+charPretty '\ETX' = "end of text (control character)"
+charPretty '\EOT' = "end of transmission (control character)"
+charPretty '\ENQ' = "enquiry (control character)"
+charPretty '\ACK' = "acknowledge (control character)"
+charPretty '\BEL' = "bell (control character)"
+charPretty '\BS'  = "backspace"
+charPretty '\t'   = "tab"
+charPretty '\n'   = "newline"
+charPretty '\v'   = "vertical tab"
+charPretty '\f'   = "form feed (control character)"
+charPretty '\r'   = "carriage return"
+charPretty '\SO'  = "shift out (control character)"
+charPretty '\SI'  = "shift in (control character)"
+charPretty '\DLE' = "data link escape (control character)"
+charPretty '\DC1' = "device control one (control character)"
+charPretty '\DC2' = "device control two (control character)"
+charPretty '\DC3' = "device control three (control character)"
+charPretty '\DC4' = "device control four (control character)"
+charPretty '\NAK' = "negative acknowledge (control character)"
+charPretty '\SYN' = "synchronous idle (control character)"
+charPretty '\ETB' = "end of transmission block (control character)"
+charPretty '\CAN' = "cancel (control character)"
+charPretty '\EM'  = "end of medium (control character)"
+charPretty '\SUB' = "substitute (control character)"
+charPretty '\ESC' = "escape (control character)"
+charPretty '\FS'  = "file separator (control character)"
+charPretty '\GS'  = "group separator (control character)"
+charPretty '\RS'  = "record separator (control character)"
+charPretty '\US'  = "unit separator (control character)"
+charPretty '\DEL' = "delete (control character)"
+charPretty ' '    = "space"
+charPretty '\160' = "non-breaking space"
 charPretty x    = "'" ++ [x] ++ "'"
 
 -- | The type class defines how to print custom data component of
@@ -223,8 +298,13 @@ instance ShowErrorComponent Dec where
                 EQ -> "equal to "
                 GT -> "greater than "
 
--- | Pretty-print 'ParseError'. Note that rendered 'String' always ends with
--- a newline.
+-- | Pretty-print 'ParseError'. The rendered 'String' always ends with a
+-- newline.
+--
+-- The function is defined as:
+--
+-- > parseErrorPretty e =
+-- >   sourcePosStackPretty (errorPos e) ++ ":\n" ++ parseErrorTextPretty e
 --
 -- @since 5.0.0
 
@@ -233,14 +313,8 @@ parseErrorPretty :: ( Ord t
                     , ShowErrorComponent e )
   => ParseError t e    -- ^ Parse error to render
   -> String            -- ^ Result of rendering
-parseErrorPretty (ParseError pos us ps xs) =
-  sourcePosStackPretty pos ++ ":\n" ++
-  if E.null us && E.null ps && E.null xs
-    then "unknown parse error\n"
-    else concat
-      [ messageItemsPretty "unexpected " us
-      , messageItemsPretty "expecting "  ps
-      , unlines (showErrorComponent <$> E.toAscList xs) ]
+parseErrorPretty e =
+  sourcePosStackPretty (errorPos e) ++ ":\n" ++ parseErrorTextPretty e
 
 -- | Pretty-print stack of source positions.
 --
@@ -271,3 +345,22 @@ orList :: NonEmpty String -> String
 orList (x:|[])  = x
 orList (x:|[y]) = x ++ " or " ++ y
 orList xs       = intercalate ", " (NE.init xs) ++ ", or " ++ NE.last xs
+
+-- | Pretty-print textual part of a 'ParseError', that is, everything except
+-- stack of source positions. The rendered staring always ends with a new
+-- line.
+--
+-- @since 5.1.0
+
+parseErrorTextPretty :: ( Ord t
+                        , ShowToken t
+                        , ShowErrorComponent e )
+  => ParseError t e    -- ^ Parse error to render
+  -> String            -- ^ Result of rendering
+parseErrorTextPretty (ParseError _ us ps xs) =
+  if E.null us && E.null ps && E.null xs
+    then "unknown parse error\n"
+    else concat
+      [ messageItemsPretty "unexpected " us
+      , messageItemsPretty "expecting "  ps
+      , unlines (showErrorComponent <$> E.toAscList xs) ]
