@@ -1,6 +1,6 @@
 -- |
 -- Module      :  Text.Megaparsec.Lexer
--- Copyright   :  © 2015–2016 Megaparsec contributors
+-- Copyright   :  © 2015–2017 Megaparsec contributors
 --                © 2007 Paolo Martini
 --                © 1999–2001 Daan Leijen
 -- License     :  FreeBSD
@@ -20,6 +20,7 @@
 
 {-# LANGUAGE CPP              #-}
 {-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE MultiWayIf       #-}
 {-# LANGUAGE TypeFamilies     #-}
 
 module Text.Megaparsec.Lexer
@@ -99,15 +100,15 @@ import Control.Applicative ((<$>), (<*), (*>), (<*>), pure)
 -- of the file).
 
 space :: MonadParsec e s m
-  => m () -- ^ A parser for a space character (e.g. 'C.spaceChar')
+  => m () -- ^ A parser for a space character (e.g. @'void' 'C.spaceChar'@)
   -> m () -- ^ A parser for a line comment (e.g. 'skipLineComment')
   -> m () -- ^ A parser for a block comment (e.g. 'skipBlockComment')
   -> m ()
 space ch line block = hidden . skipMany $ choice [ch, line, block]
 
--- | This is wrapper for lexemes. Typical usage is to supply first argument
--- (parser that consumes white space, probably defined via 'space') and use
--- the resulting function to wrap parsers for every lexeme.
+-- | This is a wrapper for lexemes. Typical usage is to supply the first
+-- argument (parser that consumes white space, probably defined via 'space')
+-- and use the resulting function to wrap parsers for every lexeme.
 --
 -- > lexeme  = L.lexeme spaceConsumer
 -- > integer = lexeme L.integer
@@ -148,7 +149,7 @@ symbol' :: (MonadParsec e s m, Token s ~ Char)
   -> m String
 symbol' spc = lexeme spc . C.string'
 
--- | Given comment prefix this function returns parser that skips line
+-- | Given comment prefix this function returns a parser that skips line
 -- comments. Note that it stops just before newline character but doesn't
 -- consume the newline. Newline is either supposed to be consumed by 'space'
 -- parser or picked up manually.
@@ -158,7 +159,7 @@ skipLineComment :: (MonadParsec e s m, Token s ~ Char)
   -> m ()
 skipLineComment prefix = p >> void (manyTill C.anyChar n)
   where p = C.string prefix
-        n = lookAhead C.newline
+        n = lookAhead (void C.newline) <|> eof
 
 -- | @skipBlockComment start end@ skips non-nested block comment starting
 -- with @start@ and ending with @end@.
@@ -218,11 +219,11 @@ incorrectIndent ord ref actual = failure E.empty E.empty (E.singleton x)
 
 -- | @indentGuard spaceConsumer ord ref@ first consumes all white space
 -- (indentation) with @spaceConsumer@ parser, then it checks column
--- position. Ordering between current indentation level and reference
+-- position. Ordering between current indentation level and the reference
 -- indentation level @ref@ should be @ord@, otherwise the parser fails. On
--- success current column position is returned.
+-- success the current column position is returned.
 --
--- When you want to parse block of indentation first run this parser with
+-- When you want to parse a block of indentation, first run this parser with
 -- arguments like @indentGuard spaceConsumer GT (unsafePos 1)@ — this will
 -- make sure you have some indentation. Use returned value to check
 -- indentation on every subsequent line according to syntax of your
@@ -240,7 +241,7 @@ indentGuard sc ord ref = do
     then return actual
     else incorrectIndent ord ref actual
 
--- | Parse non-indented construction. This ensures that there is no
+-- | Parse a non-indented construction. This ensures that there is no
 -- indentation before actual data. Useful, for example, as a wrapper for
 -- top-level function definitions.
 --
@@ -264,7 +265,7 @@ data IndentOpt m a b
     -- ^ Parse many indented tokens (possibly zero), use given indentation
     -- level (if 'Nothing', use level of the first indented token); the
     -- second argument tells how to get final result, and third argument
-    -- describes how to parse indented token
+    -- describes how to parse an indented token
   | IndentSome (Maybe Pos) ([b] -> m a) (m b)
     -- ^ Just like 'IndentMany', but requires at least one indented token to
     -- be present
@@ -289,15 +290,19 @@ indentBlock sc r = do
   ref <- indentLevel
   a   <- r
   case a of
-    IndentNone x -> return x
+    IndentNone x -> sc *> return x
     IndentMany indent f p -> do
-      mlvl <- optional . try $ C.eol *> indentGuard sc GT ref
-      case mlvl of
-        Nothing  -> sc *> f []
-        Just lvl -> indentedItems ref (fromMaybe lvl indent) sc p >>= f
+      mlvl <- (optional . try) (C.eol *> indentGuard sc GT ref)
+      done <- isJust <$> optional eof
+      case (mlvl, done) of
+        (Just lvl, False) ->
+          indentedItems ref (fromMaybe lvl indent) sc p >>= f
+        _ -> sc *> f []
     IndentSome indent f p -> do
       lvl <- C.eol *> indentGuard sc GT ref
-      indentedItems ref (fromMaybe lvl indent) sc p >>= f
+      x   <- p
+      xs  <- indentedItems ref (fromMaybe lvl indent) sc p
+      f (x:xs)
 
 -- | Grab indented items. This is a helper for 'indentBlock', it's not a
 -- part of public API.
@@ -310,15 +315,15 @@ indentedItems :: MonadParsec e s m
   -> m [b]
 indentedItems ref lvl sc p = go
   where
-    go = (sc *> indentLevel) >>= re
-    re pos
-      | pos <= ref = return []
-      | pos == lvl = (:) <$> p <*> go
-      | otherwise  = do
-          done <- isJust <$> optional eof
-          if done
-            then return []
-            else incorrectIndent EQ lvl pos
+    go = do
+      sc
+      pos  <- indentLevel
+      done <- isJust <$> optional eof
+      if done
+        then return []
+        else if | pos <= ref -> return []
+                | pos == lvl -> (:) <$> p <*> go
+                | otherwise  -> incorrectIndent EQ lvl pos
 
 -- | Create a parser that supports line-folding. The first argument is used
 -- to consume white space between components of line fold, thus it /must/
@@ -361,6 +366,14 @@ lineFold sc action =
 -- string literals:
 --
 -- > stringLiteral = char '"' >> manyTill L.charLiteral (char '"')
+--
+-- If you want to write @stringLiteral@ that adheres to the Haskell report
+-- though, you'll need to take care of the @\\&@ combination which is not a
+-- character, but can be used to separate characters (as in @\"\\291\\&4\"@
+-- which is two characters long):
+--
+-- > stringLiteral = catMaybes <$> (char '"' >> manyTill ch (char '"'))
+-- >   where ch = (Just <$> L.charLiteral) <|> (Nothing <$ string "\\&")
 
 charLiteral :: (MonadParsec e s m, Token s ~ Char) => m Char
 charLiteral = label "literal character" $ do
@@ -375,7 +388,7 @@ charLiteral = label "literal character" $ do
 -- Numbers
 
 -- | Parse an integer without sign in decimal representation (according to
--- format of integer literals described in Haskell report).
+-- the format of integer literals described in the Haskell report).
 --
 -- If you need to parse signed integers, see 'signed' combinator.
 
@@ -389,10 +402,10 @@ decimal :: (MonadParsec e s m, Token s ~ Char) => m Integer
 decimal = nump "" C.digitChar <?> "decimal integer"
 
 -- | Parse an integer in hexadecimal representation. Representation of
--- hexadecimal number is expected to be according to Haskell report except
--- for the fact that this parser doesn't parse “0x” or “0X” prefix. It is
--- responsibility of the programmer to parse correct prefix before parsing
--- the number itself.
+-- hexadecimal number is expected to be according to the Haskell report
+-- except for the fact that this parser doesn't parse “0x” or “0X” prefix.
+-- It is responsibility of the programmer to parse correct prefix before
+-- parsing the number itself.
 --
 -- For example you can make it conform to Haskell report like this:
 --
@@ -402,10 +415,10 @@ hexadecimal :: (MonadParsec e s m, Token s ~ Char) => m Integer
 hexadecimal = nump "0x" C.hexDigitChar <?> "hexadecimal integer"
 
 -- | Parse an integer in octal representation. Representation of octal
--- number is expected to be according to Haskell report except for the fact
--- that this parser doesn't parse “0o” or “0O” prefix. It is responsibility
--- of the programmer to parse correct prefix before parsing the number
--- itself.
+-- number is expected to be according to the Haskell report except for the
+-- fact that this parser doesn't parse “0o” or “0O” prefix. It is
+-- responsibility of the programmer to parse correct prefix before parsing
+-- the number itself.
 
 octal :: (MonadParsec e s m, Token s ~ Char) => m Integer
 octal = nump "0o" C.octDigitChar <?> "octal integer"
@@ -417,11 +430,11 @@ octal = nump "0o" C.octDigitChar <?> "octal integer"
 nump :: MonadParsec e s m => String -> m Char -> m Integer
 nump prefix baseDigit = read . (prefix ++) <$> some baseDigit
 
--- | Parse floating point value as 'Scientific' number. 'Scientific' is
+-- | Parse a floating point value as 'Scientific' number. 'Scientific' is
 -- great for parsing of arbitrary precision numbers coming from an untrusted
 -- source. See documentation in "Data.Scientific" for more information.
--- Representation of floating point value is expected to be according to
--- Haskell report.
+-- Representation of the floating point value is expected to be according to
+-- the Haskell report.
 --
 -- This function does not parse sign, if you need to parse signed numbers,
 -- see 'signed'.
@@ -432,7 +445,7 @@ scientific :: (MonadParsec e s m, Token s ~ Char) => m Scientific
 scientific = label "floating point number" (read <$> f)
   where f = (++) <$> some C.digitChar <*> (fraction <|> fExp)
 
--- | Parse floating point number without sign. This is a simple shortcut
+-- | Parse a floating point number without sign. This is a simple shortcut
 -- defined as:
 --
 -- > float = toRealFloat <$> scientific
@@ -468,10 +481,10 @@ number :: (MonadParsec e s m, Token s ~ Char) => m Scientific
 number = label "number" (read <$> f)
   where f = (++) <$> some C.digitChar <*> option "" (fraction <|> fExp)
 
--- | @signed space p@ parser parses optional sign, then if there is a sign
--- it will consume optional white space (using @space@ parser), then it runs
--- parser @p@ which should return a number. Sign of the number is changed
--- according to previously parsed sign.
+-- | @signed space p@ parser parses an optional sign, then if there is a
+-- sign it will consume optional white space (using @space@ parser), then it
+-- runs parser @p@ which should return a number. Sign of the number is
+-- changed according to previously parsed sign.
 --
 -- For example, to parse signed integer you can write:
 --
