@@ -1,6 +1,6 @@
 -- |
 -- Module      :  Text.Megaparsec.Pos
--- Copyright   :  © 2015–2016 Megaparsec contributors
+-- Copyright   :  © 2015–2017 Megaparsec contributors
 -- License     :  FreeBSD
 --
 -- Maintainer  :  Mark Karpov <markkarpov@opmbx.org>
@@ -39,10 +39,10 @@ import Data.Data (Data)
 import Data.Semigroup
 import Data.Typeable (Typeable)
 import GHC.Generics
-import Unsafe.Coerce
+import Test.QuickCheck
 
 #if !MIN_VERSION_base(4,8,0)
-import Control.Applicative ((<$>))
+import Control.Applicative
 import Data.Word (Word)
 #endif
 
@@ -58,6 +58,9 @@ import Data.Word (Word)
 newtype Pos = Pos Word
   deriving (Show, Eq, Ord, Data, Typeable, NFData)
 
+instance Arbitrary Pos where
+  arbitrary = unsafePos <$> (getSmall <$> arbitrary `suchThat` (> 0))
+
 -- | Construction of 'Pos' from an instance of 'Integral'. The function
 -- throws 'InvalidPosException' when given non-positive argument. Note that
 -- the function is polymorphic with respect to 'MonadThrow' @m@, so you can
@@ -89,7 +92,7 @@ unsafePos x =
 -- @since 5.0.0
 
 unPos :: Pos -> Word
-unPos = unsafeCoerce
+unPos (Pos w) = w
 {-# INLINE unPos #-}
 
 instance Semigroup Pos where
@@ -103,6 +106,14 @@ instance Read Pos where
       (x,     r3) <- readsPrec 11 r2
       (,r3) <$> mkPos (x :: Integer)
 
+instance Arbitrary SourcePos where
+  arbitrary = SourcePos
+    <$> sized (\n -> do
+          k <- choose (0, n `div` 2)
+          vectorOf k arbitrary)
+    <*> (unsafePos <$> choose (1, 1000))
+    <*> (unsafePos <$> choose (1,  100))
+
 -- | The exception is thrown by 'mkPos' when its argument is not a positive
 -- number.
 --
