@@ -1,6 +1,6 @@
 -- |
 -- Module      :  Text.Megaparsec.Prim
--- Copyright   :  © 2015–2016 Megaparsec contributors
+-- Copyright   :  © 2015–2017 Megaparsec contributors
 --                © 2007 Paolo Martini
 --                © 1999–2001 Daan Leijen
 -- License     :  FreeBSD
@@ -44,17 +44,21 @@ module Text.Megaparsec.Prim
   , setPosition
   , pushPosition
   , popPosition
+  , getTokensProcessed
+  , setTokensProcessed
   , getTabWidth
   , setTabWidth
   , setParserState
     -- * Running parser
+  , parse
+  , parseMaybe
+  , parseTest
   , runParser
   , runParser'
   , runParserT
   , runParserT'
-  , parse
-  , parseMaybe
-  , parseTest )
+    -- * Debugging
+  , dbg )
 where
 
 import Control.DeepSeq
@@ -68,16 +72,21 @@ import Control.Monad.Trans
 import Control.Monad.Trans.Identity
 import Data.Data (Data)
 import Data.Foldable (foldl')
+import Data.List (genericTake)
 import Data.List.NonEmpty (NonEmpty (..))
 import Data.Monoid hiding ((<>))
 import Data.Proxy
 import Data.Semigroup
 import Data.Set (Set)
 import Data.Typeable (Typeable)
+import Debug.Trace
 import GHC.Generics
 import Prelude hiding (all)
+import Test.QuickCheck hiding (Result (..), label)
 import qualified Control.Applicative               as A
 import qualified Control.Monad.Fail                as Fail
+import qualified Control.Monad.RWS.Lazy            as L
+import qualified Control.Monad.RWS.Strict          as S
 import qualified Control.Monad.Trans.Reader        as L
 import qualified Control.Monad.Trans.State.Lazy    as L
 import qualified Control.Monad.Trans.State.Strict  as S
@@ -94,7 +103,8 @@ import Text.Megaparsec.Error
 import Text.Megaparsec.Pos
 
 #if !MIN_VERSION_base(4,8,0)
-import Control.Applicative ((<$>), (<*), pure)
+import Control.Applicative
+import Data.Word (Word)
 #endif
 
 ----------------------------------------------------------------------------
@@ -103,13 +113,32 @@ import Control.Applicative ((<$>), (<*), pure)
 -- | This is Megaparsec's state, it's parametrized over stream type @s@.
 
 data State s = State
-  { stateInput    :: s
-  , statePos      :: NonEmpty SourcePos
-  , stateTabWidth :: Pos }
-  deriving (Show, Eq, Data, Typeable, Generic)
+  { stateInput :: s
+    -- ^ Current input (already processed input is removed from the stream)
+  , statePos :: NonEmpty SourcePos
+    -- ^ Current position (column + line number) with support for include files
+  , stateTokensProcessed :: {-# UNPACK #-} !Word
+    -- ^ Number of processed tokens so far
+    --
+    -- @since 5.2.0
+  , stateTabWidth :: Pos
+    -- ^ Tab width to use
+  } deriving (Show, Eq, Data, Typeable, Generic)
 
 instance NFData s => NFData (State s)
 
+instance Arbitrary a => Arbitrary (State a) where
+  arbitrary = State
+    <$> arbitrary
+    <*>
+#if !MIN_VERSION_QuickCheck(2,9,0)
+      (NE.fromList . getNonEmpty <$> arbitrary)
+#else
+      arbitrary
+#endif
+    <*> choose (1, 10000)
+    <*> (unsafePos <$> choose (1, 20))
+
 -- | All information available after parsing. This includes consumption of
 -- input, success (with returned value) or failure (with parse error), and
 -- parser state at the end of parsing.
@@ -359,23 +388,6 @@ pAp m k = ParsecT $ \s cok cerr eok eerr ->
 instance (ErrorComponent e, Stream s) => A.Alternative (ParsecT e s m) where
   empty  = mzero
   (<|>)  = mplus
-  many p = reverse <$> manyAcc p
-
-manyAcc :: ParsecT e s m a -> ParsecT e s m [a]
-manyAcc p = ParsecT $ \s cok cerr eok _ ->
-  let errToHints c err _ = c (toHints err)
-      walk xs x s' _ =
-        unParser p s'
-        (seq xs $ walk $ x:xs)       -- consumed-OK
-        cerr                         -- consumed-error
-        manyErr                      -- empty-OK
-        (errToHints $ cok (x:xs) s') -- empty-error
-  in unParser p s (walk []) cerr manyErr (errToHints $ eok [] s)
-
-manyErr :: a
-manyErr = error $
-  "Text.Megaparsec.Prim.many: combinator 'many' is applied to a parser"
-  ++ " that may consume no input."
 
 instance (ErrorComponent e, Stream s)
     => Monad (ParsecT e s m) where
@@ -404,7 +416,7 @@ instance (ErrorComponent e, Stream s)
   fail = pFail
 
 pFail :: ErrorComponent e => String -> ParsecT e s m a
-pFail msg = ParsecT $ \s@(State _ pos _) _ _ _ eerr ->
+pFail msg = ParsecT $ \s@(State _ pos _ _) _ _ _ eerr ->
   eerr (ParseError pos E.empty E.empty d) s
   where d = E.singleton (representFail msg)
 {-# INLINE pFail #-}
@@ -458,7 +470,7 @@ instance (ErrorComponent e, Stream s)
   mplus = pPlus
 
 pZero :: ParsecT e s m a
-pZero = ParsecT $ \s@(State _ pos _) _ _ _ eerr ->
+pZero = ParsecT $ \s@(State _ pos _ _) _ _ _ eerr ->
   eerr (ParseError pos E.empty E.empty E.empty) s
 {-# INLINE pZero #-}
 
@@ -475,12 +487,13 @@ pPlus m n = ParsecT $ \s cok cerr eok eerr ->
   in unParser m s cok cerr eok meerr
 {-# INLINE pPlus #-}
 
--- | From two states, return the one with greater textual position. If the
--- positions are equal, prefer the latter state.
+-- | From two states, return the one with greater number of processed
+-- tokens. If the numbers of processed tokens are equal, prefer the latter
+-- state.
 
 longestMatch :: State s -> State s -> State s
-longestMatch s1@(State _ pos1 _) s2@(State _ pos2 _) =
-  case pos1 `compare` pos2 of
+longestMatch s1@(State _ _ tp1 _) s2@(State _ _ tp2 _) =
+  case tp1 `compare` tp2 of
     LT -> s2
     EQ -> s2
     GT -> s1
@@ -524,8 +537,8 @@ class (ErrorComponent e, Stream s, A.Alternative m, MonadPlus m)
   hidden :: m a -> m a
   hidden = label ""
 
-  -- | The parser @try p@ behaves like parser @p@, except that it
-  -- pretends that it hasn't consumed any input when an error occurs.
+  -- | The parser @try p@ behaves like parser @p@, except that it backtracks
+  -- parser state when @p@ fails (either consuming input or not).
   --
   -- This combinator is used whenever arbitrary look ahead is needed. Since
   -- it pretends that it hasn't consumed any input when @p@ fails, the
@@ -563,16 +576,17 @@ class (ErrorComponent e, Stream s, A.Alternative m, MonadPlus m)
 
   try :: m a -> m a
 
-  -- | @lookAhead p@ parses @p@ without consuming any input.
-  --
-  -- If @p@ fails and consumes some input, so does @lookAhead@. Combine with
-  -- 'try' if this is undesirable.
+  -- | If @p@ in @lookAhead p@ succeeds (either consuming input or not) the
+  -- whole parser behaves like @p@ succeeded without consuming anything
+  -- (parser state is not updated as well). If @p@ fails, @lookAhead@ has no
+  -- effect, i.e. it will fail consuming input if @p@ fails consuming input.
+  -- Combine with 'try' if this is undesirable.
 
   lookAhead :: m a -> m a
 
   -- | @notFollowedBy p@ only succeeds when parser @p@ fails. This parser
-  -- does not consume any input and can be used to implement the “longest
-  -- match” rule.
+  -- /never consumes/ any input and /never modifies/ parser state. It can be
+  -- used to implement the “longest match” rule.
 
   notFollowedBy :: m a -> m ()
 
@@ -593,6 +607,19 @@ class (ErrorComponent e, Stream s, A.Alternative m, MonadPlus m)
     -> m a             -- ^ Original parser
     -> m a             -- ^ Parser that can recover from failures
 
+  -- | @observing p@ allows to “observe” failure of @p@ parser, should it
+  -- happen, without actually ending parsing, but instead getting the
+  -- 'ParseError' in 'Left'. On success parsed value is returned in 'Right'
+  -- as usual. Note that this primitive just allows you to observe parse
+  -- errors as they happen, it does not backtrack or change how the @p@
+  -- parser works in any way.
+  --
+  -- @since 5.1.0
+
+  observing
+    :: m a
+    -> m (Either (ParseError (Token s) e) a)
+
   -- | This parser only succeeds at the end of the input.
 
   eof :: m ()
@@ -668,6 +695,7 @@ instance (ErrorComponent e, Stream s) => MonadParsec e s (ParsecT e s m) where
   lookAhead         = pLookAhead
   notFollowedBy     = pNotFollowedBy
   withRecovery      = pWithRecovery
+  observing         = pObserving
   eof               = pEof
   token             = pToken
   tokens            = pTokens
@@ -679,7 +707,7 @@ pFailure
   -> Set (ErrorItem (Token s))
   -> Set e
   -> ParsecT e s m a
-pFailure us ps xs = ParsecT $ \s@(State _ pos _) _ _ _ eerr ->
+pFailure us ps xs = ParsecT $ \s@(State _ pos _ _) _ _ _ eerr ->
   eerr (ParseError pos us ps xs) s
 {-# INLINE pFailure #-}
 
@@ -696,7 +724,8 @@ pLabel l p = ParsecT $ \s cok cerr eok eerr ->
 
 pTry :: ParsecT e s m a -> ParsecT e s m a
 pTry p = ParsecT $ \s cok _ eok eerr ->
-  unParser p s cok eerr eok eerr
+  let eerr' err _ = eerr err s
+  in unParser p s cok eerr' eok eerr'
 {-# INLINE pTry #-}
 
 pLookAhead :: ParsecT e s m a -> ParsecT e s m a
@@ -706,7 +735,7 @@ pLookAhead p = ParsecT $ \s _ cerr eok eerr ->
 {-# INLINE pLookAhead #-}
 
 pNotFollowedBy :: Stream s => ParsecT e s m a -> ParsecT e s m ()
-pNotFollowedBy p = ParsecT $ \s@(State input pos _) _ _ eok eerr ->
+pNotFollowedBy p = ParsecT $ \s@(State input pos _ _) _ _ eok eerr ->
   let what = maybe EndOfInput (Tokens . nes . fst) (uncons input)
       unexpect u = ParseError pos (E.singleton u) E.empty E.empty
       cok' _ _ _ = eerr (unexpect what) s
@@ -736,8 +765,17 @@ pWithRecovery r p = ParsecT $ \s cok cerr eok eerr ->
   in unParser p s cok mcerr eok meerr
 {-# INLINE pWithRecovery #-}
 
+pObserving
+  :: ParsecT e s m a
+  -> ParsecT e s m (Either (ParseError (Token s) e) a)
+pObserving p = ParsecT $ \s cok _ eok _ ->
+  let cerr' err s' = cok (Left err) s' mempty
+      eerr' err s' = eok (Left err) s' (toHints err)
+  in unParser p s (cok . Right) cerr' (eok . Right) eerr'
+{-# INLINE pObserving #-}
+
 pEof :: forall e s m. Stream s => ParsecT e s m ()
-pEof = ParsecT $ \s@(State input (pos:|z) w) _ _ eok eerr ->
+pEof = ParsecT $ \s@(State input (pos:|z) tp w) _ _ eok eerr ->
   case uncons input of
     Nothing    -> eok () s mempty
     Just (x,_) ->
@@ -747,7 +785,7 @@ pEof = ParsecT $ \s@(State input (pos:|z) w) _ _ eok eerr ->
           , errorUnexpected = (E.singleton . Tokens . nes) x
           , errorExpected   = E.singleton EndOfInput
           , errorCustom     = E.empty }
-          (State input (apos:|z) w)
+          (State input (apos:|z) tp w)
 {-# INLINE pEof #-}
 
 pToken :: forall e s m a. Stream s
@@ -756,7 +794,7 @@ pToken :: forall e s m a. Stream s
                         , Set e ) a)
   -> Maybe (Token s)
   -> ParsecT e s m a
-pToken test mtoken = ParsecT $ \s@(State input (pos:|z) w) cok _ _ eerr ->
+pToken test mtoken = ParsecT $ \s@(State input (pos:|z) tp w) cok _ _ eerr ->
   case uncons input of
     Nothing -> eerr ParseError
       { errorPos        = pos:|z
@@ -769,9 +807,9 @@ pToken test mtoken = ParsecT $ \s@(State input (pos:|z) w) cok _ _ eerr ->
         Left (us, ps, xs) ->
           apos `seq` eerr
             (ParseError (apos:|z) us ps xs)
-            (State input (apos:|z) w)
+            (State input (apos:|z) tp w)
         Right x ->
-          let newstate = State cs (npos:|z) w
+          let newstate = State cs (npos:|z) (tp + 1) w
           in npos `seq` cok x newstate mempty
 {-# INLINE pToken #-}
 
@@ -780,7 +818,7 @@ pTokens :: forall e s m. Stream s
   -> [Token s]
   -> ParsecT e s m [Token s]
 pTokens _ [] = ParsecT $ \s _ _ eok _ -> eok [] s mempty
-pTokens test tts = ParsecT $ \s@(State input (pos:|z) w) cok _ _ eerr ->
+pTokens test tts = ParsecT $ \s@(State input (pos:|z) tp w) cok _ _ eerr ->
   let updatePos' = updatePos (Proxy :: Proxy s) w
       toTokens   = Tokens . NE.fromList . reverse
       unexpect pos' u = ParseError
@@ -790,20 +828,23 @@ pTokens test tts = ParsecT $ \s@(State input (pos:|z) w) cok _ _ eerr ->
         , errorCustom     = E.empty }
       go _ [] is rs =
         let ris   = reverse is
-            !npos = foldl' (\p t -> snd (updatePos' p t)) pos ris
-        in cok ris (State rs (npos:|z) w) mempty
+            (npos, tp') = foldl'
+              (\(p, n) t -> (snd (updatePos' p t), n + 1))
+              (pos, tp)
+              ris
+        in cok ris (State rs (npos:|z) tp' w) mempty
       go apos (t:ts) is rs =
         case uncons rs of
           Nothing ->
             apos `seq` eerr
               (unexpect (apos:|z) (toTokens is))
-              (State input (apos:|z) w)
+              (State input (apos:|z) tp w)
           Just (x,xs) ->
             if test t x
               then go apos ts (x:is) xs
               else apos `seq` eerr
                      (unexpect (apos:|z) . toTokens $ x:is)
-                     (State input (apos:|z) w)
+                     (State input (apos:|z) tp w)
   in case uncons input of
        Nothing ->
          eerr (unexpect (pos:|z) EndOfInput) s
@@ -814,7 +855,7 @@ pTokens test tts = ParsecT $ \s@(State input (pos:|z) w) cok _ _ eerr ->
               then go apos ts [x] xs
               else apos `seq` eerr
                      (unexpect (apos:|z) $ Tokens (nes x))
-                     (State input (apos:|z) w)
+                     (State input (apos:|z) tp w)
 {-# INLINE pTokens #-}
 
 pGetParserState :: ParsecT e s m (State s)
@@ -857,7 +898,7 @@ getInput = stateInput <$> getParserState
 -- 'setInput' functions can for example be used to deal with include files.
 
 setInput :: MonadParsec e s m => s -> m ()
-setInput s = updateParserState (\(State _ pos w) -> State s pos w)
+setInput s = updateParserState (\(State _ pos tp w) -> State s pos tp w)
 
 -- | Return the current source position.
 --
@@ -871,8 +912,8 @@ getPosition = NE.head . statePos <$> getParserState
 -- See also: 'getPosition', 'pushPosition', 'popPosition', and 'SourcePos'.
 
 setPosition :: MonadParsec e s m => SourcePos -> m ()
-setPosition pos = updateParserState $ \(State s (_:|z) w) ->
-  State s (pos:|z) w
+setPosition pos = updateParserState $ \(State s (_:|z) tp w) ->
+  State s (pos:|z) tp w
 
 -- | Push given position into stack of positions and continue parsing
 -- working with this position. Useful for working with include files and the
@@ -883,8 +924,8 @@ setPosition pos = updateParserState $ \(State s (_:|z) w) ->
 -- @since 5.0.0
 
 pushPosition :: MonadParsec e s m => SourcePos -> m ()
-pushPosition pos = updateParserState $ \(State s z w) ->
-  State s (NE.cons pos z) w
+pushPosition pos = updateParserState $ \(State s z tp w) ->
+  State s (NE.cons pos z) tp w
 
 -- | Pop a position from stack of positions unless it only contains one
 -- element (in that case stack of positions remains the same). This is how
@@ -895,10 +936,25 @@ pushPosition pos = updateParserState $ \(State s z w) ->
 -- @since 5.0.0
 
 popPosition :: MonadParsec e s m => m ()
-popPosition = updateParserState $ \(State s z w) ->
+popPosition = updateParserState $ \(State s z tp w) ->
   case snd (NE.uncons z) of
-    Nothing -> State s z w
-    Just z' -> State s z' w
+    Nothing -> State s z  tp w
+    Just z' -> State s z' tp w
+
+-- | Get number of tokens processed so far.
+--
+-- @since 5.2.0
+
+getTokensProcessed :: MonadParsec e s m => m Word
+getTokensProcessed = stateTokensProcessed <$> getParserState
+
+-- | Set number of tokens processed so far.
+--
+-- @since 5.2.0
+
+setTokensProcessed :: MonadParsec e s m => Word -> m ()
+setTokensProcessed tp = updateParserState $ \(State s pos _ w) ->
+  State s pos tp w
 
 -- | Return tab width. Default tab width is equal to 'defaultTabWidth'. You
 -- can set different tab width with help of 'setTabWidth'.
@@ -910,7 +966,8 @@ getTabWidth = stateTabWidth <$> getParserState
 -- 'defaultTabWidth' will be used.
 
 setTabWidth :: MonadParsec e s m => Pos -> m ()
-setTabWidth w = updateParserState (\(State s pos _) -> State s pos w)
+setTabWidth w = updateParserState $ \(State s pos tp _) ->
+  State s pos tp w
 
 -- | @setParserState st@ set the full parser state to @st@.
 
@@ -1027,11 +1084,6 @@ runParserT' p s = do
     OK    x -> return (s', Right x)
     Error e -> return (s', Left  e)
 
--- | Given name of source file and input construct initial state for parser.
-
-initialState :: String -> s -> State s
-initialState name s = State s (initialPos name :| []) defaultTabWidth
-
 -- | Low-level unpacking of the 'ParsecT' type. 'runParserT' and 'runParser'
 -- are built upon this.
 
@@ -1045,6 +1097,15 @@ runParsecT p s = unParser p s cok cerr eok eerr
         eok a s' _  = return $ Reply s' Virgin   (OK a)
         eerr err s' = return $ Reply s' Virgin   (Error err)
 
+-- | Given name of source file and input construct initial state for parser.
+
+initialState :: String -> s -> State s
+initialState name s = State
+  { stateInput           = s
+  , statePos             = initialPos name :| []
+  , stateTokensProcessed = 0
+  , stateTabWidth        = defaultTabWidth }
+
 ----------------------------------------------------------------------------
 -- Instances of 'MonadParsec'
 
@@ -1058,6 +1119,8 @@ instance MonadParsec e s m => MonadParsec e s (L.StateT st m) where
     notFollowedBy (fst <$> m s) >> return ((),s)
   withRecovery r (L.StateT m) = L.StateT $ \s ->
     withRecovery (\e -> L.runStateT (r e) s) (m s)
+  observing     (L.StateT m) = L.StateT $ \s ->
+    fixs s <$> observing (m s)
   eof                        = lift eof
   token test mt              = lift (token test mt)
   tokens e ts                = lift (tokens e ts)
@@ -1074,13 +1137,15 @@ instance MonadParsec e s m => MonadParsec e s (S.StateT st m) where
     notFollowedBy (fst <$> m s) >> return ((),s)
   withRecovery r (S.StateT m) = S.StateT $ \s ->
     withRecovery (\e -> S.runStateT (r e) s) (m s)
+  observing     (S.StateT m) = S.StateT $ \s ->
+    fixs s <$> observing (m s)
   eof                        = lift eof
   token test mt              = lift (token test mt)
   tokens e ts                = lift (tokens e ts)
   getParserState             = lift getParserState
   updateParserState f        = lift (updateParserState f)
 
-instance MonadParsec e s m => MonadParsec e s (L.ReaderT st m) where
+instance MonadParsec e s m => MonadParsec e s (L.ReaderT r m) where
   failure us ps xs            = lift (failure us ps xs)
   label n       (L.ReaderT m) = L.ReaderT $ label n . m
   try           (L.ReaderT m) = L.ReaderT $ try . m
@@ -1088,6 +1153,7 @@ instance MonadParsec e s m => MonadParsec e s (L.ReaderT st m) where
   notFollowedBy (L.ReaderT m) = L.ReaderT $ notFollowedBy . m
   withRecovery r (L.ReaderT m) = L.ReaderT $ \s ->
     withRecovery (\e -> L.runReaderT (r e) s) (m s)
+  observing     (L.ReaderT m) = L.ReaderT $ observing . m
   eof                         = lift eof
   token test mt               = lift (token test mt)
   tokens e ts                 = lift (tokens e ts)
@@ -1104,6 +1170,8 @@ instance (Monoid w, MonadParsec e s m) => MonadParsec e s (L.WriterT w m) where
     (,mempty) <$> notFollowedBy (fst <$> m)
   withRecovery r (L.WriterT m) = L.WriterT $
     withRecovery (L.runWriterT . r) m
+  observing     (L.WriterT m) = L.WriterT $
+    fixs mempty <$> observing m
   eof                         = lift eof
   token test mt               = lift (token test mt)
   tokens e ts                 = lift (tokens e ts)
@@ -1120,6 +1188,48 @@ instance (Monoid w, MonadParsec e s m) => MonadParsec e s (S.WriterT w m) where
     (,mempty) <$> notFollowedBy (fst <$> m)
   withRecovery r (S.WriterT m) = S.WriterT $
     withRecovery (S.runWriterT . r) m
+  observing     (S.WriterT m) = S.WriterT $
+    fixs mempty <$> observing m
+  eof                         = lift eof
+  token test mt               = lift (token test mt)
+  tokens e ts                 = lift (tokens e ts)
+  getParserState              = lift getParserState
+  updateParserState f         = lift (updateParserState f)
+
+instance (Monoid w, MonadParsec e s m) => MonadParsec e s (L.RWST r w st m) where
+  failure us ps xs            = lift (failure us ps xs)
+  label n          (L.RWST m) = L.RWST $ \r s -> label n (m r s)
+  try              (L.RWST m) = L.RWST $ \r s -> try (m r s)
+  lookAhead        (L.RWST m) = L.RWST $ \r s -> do
+    (x,_,_) <- lookAhead (m r s)
+    return (x,s,mempty)
+  notFollowedBy    (L.RWST m) = L.RWST $ \r s -> do
+    notFollowedBy (void $ m r s)
+    return ((),s,mempty)
+  withRecovery   n (L.RWST m) = L.RWST $ \r s ->
+    withRecovery (\e -> L.runRWST (n e) r s) (m r s)
+  observing        (L.RWST m) = L.RWST $ \r s ->
+    fixs' s <$> observing (m r s)
+  eof                         = lift eof
+  token test mt               = lift (token test mt)
+  tokens e ts                 = lift (tokens e ts)
+  getParserState              = lift getParserState
+  updateParserState f         = lift (updateParserState f)
+
+instance (Monoid w, MonadParsec e s m) => MonadParsec e s (S.RWST r w st m) where
+  failure us ps xs            = lift (failure us ps xs)
+  label n          (S.RWST m) = S.RWST $ \r s -> label n (m r s)
+  try              (S.RWST m) = S.RWST $ \r s -> try (m r s)
+  lookAhead        (S.RWST m) = S.RWST $ \r s -> do
+    (x,_,_) <- lookAhead (m r s)
+    return (x,s,mempty)
+  notFollowedBy    (S.RWST m) = S.RWST $ \r s -> do
+    notFollowedBy (void $ m r s)
+    return ((),s,mempty)
+  withRecovery   n (S.RWST m) = S.RWST $ \r s ->
+    withRecovery (\e -> S.runRWST (n e) r s) (m r s)
+  observing        (S.RWST m) = S.RWST $ \r s ->
+    fixs' s <$> observing (m r s)
   eof                         = lift eof
   token test mt               = lift (token test mt)
   tokens e ts                 = lift (tokens e ts)
@@ -1134,8 +1244,138 @@ instance MonadParsec e s m => MonadParsec e s (IdentityT m) where
   notFollowedBy (IdentityT m) = IdentityT $ notFollowedBy m
   withRecovery r (IdentityT m) = IdentityT $
     withRecovery (runIdentityT . r) m
+  observing     (IdentityT m) = IdentityT $ observing m
   eof                         = lift eof
   token test mt               = lift (token test mt)
   tokens e ts                 = lift $ tokens e ts
   getParserState              = lift getParserState
   updateParserState f         = lift $ updateParserState f
+
+fixs :: s -> Either a (b, s) -> (Either a b, s)
+fixs s (Left a)       = (Left a, s)
+fixs _ (Right (b, s)) = (Right b, s)
+{-# INLINE fixs #-}
+
+fixs' :: Monoid w => s -> Either a (b, s, w) -> (Either a b, s, w)
+fixs' s (Left a)        = (Left a, s, mempty)
+fixs' _ (Right (b,s,w)) = (Right b, s, w)
+{-# INLINE fixs' #-}
+
+----------------------------------------------------------------------------
+-- Debugging
+
+-- | @dbg label p@ parser works exactly like @p@, but when it's evaluated it
+-- prints information useful for debugging. The @label@ is only used to
+-- refer to this parser in the debugging output. This combinator uses the
+-- 'trace' function from "Debug.Trace" under the hood.
+--
+-- Typical usage is to wrap every sub-parser in misbehaving parser with
+-- 'dbg' assigning meaningful labels. Then give it a shot and go through the
+-- print-out. As of current version, this combinator prints all available
+-- information except for /hints/, which are probably only interesting to
+-- the maintainer of Megaparsec itself and may be quite verbose to output in
+-- general. Let me know if you would like to be able to see hints as part of
+-- debugging output.
+--
+-- The output itself is pretty self-explanatory, although the following
+-- abbreviations should be clarified (they are derived from low-level source
+-- code):
+--
+--     * @COK@ — “consumed OK”. The parser consumed input and succeeded.
+--     * @CERR@ — “consumed error”. The parser consumed input and failed.
+--     * @EOK@ — “empty OK”. The parser succeeded without consuming input.
+--     * @EERR@ — “empty error”. The parser failed without consuming input.
+--
+-- Finally, it's not possible to lift this function into some monad
+-- transformers without introducing surprising behavior (e.g. unexpected
+-- state backtracking) or adding otherwise redundant constraints (e.g.
+-- 'Show' instance for state), so this helper is only available for
+-- 'ParsecT' monad, not 'MonadParsec' in general.
+--
+-- @since 5.1.0
+
+dbg :: forall e s m a.
+  ( Stream s
+  , ShowToken (Token s)
+  , ShowErrorComponent e
+  , Show a )
+  => String            -- ^ Debugging label
+  -> ParsecT e s m a   -- ^ Parser to debug
+  -> ParsecT e s m a   -- ^ Parser that prints debugging messages
+dbg lbl p = ParsecT $ \s cok cerr eok eerr ->
+  let l = dbgLog lbl :: DbgItem s e a -> String
+      cok' x s' hs = flip trace (cok x s' hs) $
+        l (DbgIn (unfold (stateInput s))) ++
+        l (DbgCOK (streamTake (streamDelta s s') (stateInput s)) x)
+      cerr' err s' = flip trace (cerr err s') $
+        l (DbgIn (unfold (stateInput s))) ++
+        l (DbgCERR (streamTake (streamDelta s s') (stateInput s)) err)
+      eok' x s' hs = flip trace (eok x s' hs) $
+        l (DbgIn (unfold (stateInput s))) ++
+        l (DbgEOK (streamTake (streamDelta s s') (stateInput s)) x)
+      eerr' err s' = flip trace (eerr err s') $
+        l (DbgIn (unfold (stateInput s))) ++
+        l (DbgEERR (streamTake (streamDelta s s') (stateInput s)) err)
+  in unParser p s cok' cerr' eok' eerr'
+
+-- | Single piece of info to be rendered with 'dbgLog'.
+
+data DbgItem s e a
+  = DbgIn   [Token s]
+  | DbgCOK  [Token s] a
+  | DbgCERR [Token s] (ParseError (Token s) e)
+  | DbgEOK  [Token s] a
+  | DbgEERR [Token s] (ParseError (Token s) e)
+
+-- | Render a single piece of debugging info.
+
+dbgLog :: (ShowToken (Token s), ShowErrorComponent e, Show a, Ord (Token s))
+  => String            -- ^ Debugging label
+  -> DbgItem s e a     -- ^ Information to render
+  -> String            -- ^ Rendered result
+dbgLog lbl item = prefix msg
+  where
+    prefix = unlines . fmap ((lbl ++ "> ") ++) . lines
+    msg = case item of
+      DbgIn   ts   ->
+        "IN: " ++ showStream ts
+      DbgCOK  ts a ->
+        "MATCH (COK): " ++ showStream ts ++ "\nVALUE: " ++ show a
+      DbgCERR ts e ->
+        "MATCH (CERR): " ++ showStream ts ++ "\nERROR:\n" ++ parseErrorPretty e
+      DbgEOK  ts a ->
+        "MATCH (EOK): " ++ showStream ts ++ "\nVALUE: " ++ show a
+      DbgEERR ts e ->
+        "MATCH (EERR): " ++ showStream ts ++ "\nERROR:\n" ++ parseErrorPretty e
+
+-- | Pretty-print a list of tokens.
+
+showStream :: ShowToken t => [t] -> String
+showStream ts =
+  case NE.nonEmpty ts of
+    Nothing -> "<EMPTY>"
+    Just ne ->
+      let (h, r) = splitAt 40 (showTokens ne)
+      in if null r then h else h ++ " <…>"
+
+-- | Calculate number of consumed tokens given 'State' of parser before and
+-- after parsing.
+
+streamDelta
+  :: State s           -- ^ State of parser before consumption
+  -> State s           -- ^ State of parser after consumption
+  -> Word              -- ^ Number of consumed tokens
+streamDelta s0 s1 = stateTokensProcessed s1 - stateTokensProcessed s0
+
+-- | Extract given number of tokens from the stream.
+
+streamTake :: Stream s => Word -> s -> [Token s]
+streamTake n s = genericTake n (unfold s)
+
+-- | Custom version of 'unfold' that matches signature of 'uncons' method in
+-- 'Stream' type class we use.
+
+unfold :: Stream s => s -> [Token s]
+unfold s = case uncons s of
+  Nothing -> []
+  Just (t, s') -> t : unfold s'
